﻿Imports System.ComponentModel
Imports System.Runtime.CompilerServices
Imports System.Reflection

Namespace System.Windows.Forms

    Public Module BindingSourceX

        ''' <summary>
        ''' provides a placeholder-syntax for filters. Sample: bs.FilterX("Datum >= ? And ? >= Datum", dtpVon.Value, dtpBis.Value)
        ''' </summary>
        ''' <exception cref="ArgumentException">expression contains more placeholders than values are passed</exception>
        <Extension()>
        Public Sub FilterX(ByVal bs As BindingSource, ByVal expression As String, ByVal ParamArray values() As Object)
            Dim splits = expression.Split("?"c)      ' identify placeholder '?'
            If splits.Count > values.Length + 1 Then Throw New ArgumentException(
               "expression contains more placeholders than values are passed")
            For i = 0 To splits.Length - 2
                Dim val = values(i)
                Dim format = If(TypeOf val Is String, "'{0}'", If(TypeOf val Is Date, "#{0}#", "{0}"))
                splits(i) &= String.Format(Globalization.CultureInfo.InvariantCulture, format, val)
            Next
            bs.Filter = String.Concat(splits)
        End Sub

        ''' <summary> editiert bs.AddNew im angegebenen Dialog-Form. </summary>
        <Extension()>
        Public Function EditNew(Of T As {Form, New})(ByVal bs As BindingSource, defaultValueColumn As DataColumn, defaultValue As Object, Optional onErrorRetry As Boolean = False, Optional owner As Form = Nothing) As DialogResult
            Return bs.EditNew(Of T)({Tuple.Create(defaultValueColumn, defaultValue)}, onErrorRetry, owner)
        End Function
        ''' <summary> editiert bs.AddNew im angegebenen Dialog-Form. </summary>
        <Extension()>
        Public Function EditNew(Of T As {Form, New})(ByVal bs As BindingSource, defaultValues As IEnumerable(Of Tuple(Of DataColumn, Object)), Optional onErrorRetry As Boolean = False, Optional owner As Form = Nothing) As DialogResult
            Dim drv = DirectCast(bs.AddNew, DataRowView)
            defaultValues.ForEach(Sub(tpl) drv(tpl.Item1.ColumnName) = tpl.Item2)
            Return EditItem(Of T)(bs, drv, onErrorRetry, owner)
        End Function
        ''' <summary> editiert bs.AddNew im angegebenen Dialog-Form. </summary>
        <Extension()>
        Public Function EditNew(Of TEditForm As {Form, New})(bs As BindingSource, Optional onErrorRetry As Boolean = False, Optional owner As Form = Nothing) As DialogResult
            Return EditItem(Of TEditForm)(bs, bs.AddNew, onErrorRetry, owner)
        End Function
        ''' <summary> editiert bs.Current im angegebenen Dialog-Form. </summary>
        <Extension()>
        Public Function EditCurrent(Of TEditForm As {Form, New})(bs As BindingSource, Optional onErrorRetry As Boolean = False, Optional owner As Form = Nothing) As DialogResult
            Return EditItem(Of TEditForm)(bs, bs.Current, onErrorRetry, owner)
        End Function
        Private Function EditItem(Of T As {Form, New})(bs As BindingSource, item As Object, onErrorRetry As Boolean, owner As Form) As DialogResult
            Dim tb = DirectCast(item, DataRowView).Row.Table
            EditItem = DialogResult.Retry
            While EditItem = DialogResult.Retry
                Using frm = New T
                    frm.Register(tb.DataSet)
                    Dim allCtls = frm.GetAllControls
                    Dim bindFlags As BindingFlags = BindingFlags.Instance Or BindingFlags.Public _
                       Or BindingFlags.NonPublic Or BindingFlags.GetField
                    For Each ctl In allCtls.Where(Function(c) TypeOf c Is ContainerControl AndAlso TypeOf c Is Form OrElse TypeOf c Is UserControl)
                        For Each fld In ctl.GetType.GetFields(bindFlags).Where(Function(f) f.FieldType Is GetType(BindingSource))
                            Dim bs2 = DirectCast(fld.GetValue(ctl), BindingSource)
                            If bs2 Is Nothing Then Continue For
                            If tb Is bs2.DataTable Then
                                bs2.DataSource = item
                                Try
                                    EditItem = frm.ShowDialog(owner)
                                    If EditItem <> DialogResult.Cancel Then
                                        Try
                                            bs2.EndEdit()
                                        Catch ex As Exception
                                            If Not onErrorRetry Then Throw
                                            EditItem = MessageBox.Show(ex.Message, ex.GetType.Name, MessageBoxButtons.RetryCancel)
                                            Continue While
                                        End Try
                                    End If
                                Finally
                                    If EditItem = DialogResult.Cancel Then
                                        bs.CancelEdit()
                                    Else
                                        If DirectCast(bs.List, DataView).Sort <> "" Then bs.MoveToRow(DirectCast(item, DataRowView).Row)
                                        bs.ResetCurrentItem() 'den von der anneren BS geänderten Datensatz neu einlesen.
                                    End If
                                End Try
                                Exit Function
                            End If
                        Next
                    Next
                    Throw New Exception("could not find a proper BindingSource.")
                End Using

            End While
        End Function

        <Extension()>
        Public Function [Dataset](bs As BindingSource) As DataSet
            Dim BindSource As BindingSource = Nothing
            Do Until bs Is Nothing
                BindSource = bs
                bs = TryCast(bs.DataSource, BindingSource)
            Loop
            Dim src = BindSource.DataSource
            Dim ds = TryCast(src, DataSet)
            If ds.NotNull Then Return ds
            Dim tb = TryCast(src, DataTable)
            If tb.NotNull Then Return tb.DataSet
            Dim rw = TryCast(src, DataRow)
            If rw.NotNull Then Return rw.Table.DataSet
            Dim drv = TryCast(src, DataRowView)
            If drv.NotNull Then Return drv.Row.Table.DataSet
            If TypeOf src Is DataView Then Throw New Exception(
               "DatasetFromBindingSource: Dont set BindingSources DataSource on DataView. It's buggy!")
            Return Nothing
        End Function

        <Extension()>
        Public Function [DataTable](Source As BindingSource) As DataTable
            While Source.DataMember = ""
                Dim drv = TryCast(Source.DataSource, DataRowView)
                If drv.NotNull Then Return drv.Row.Table
                Source = TryCast(Source.DataSource, BindingSource)
                If Source.Null Then Return Nothing
            End While
            Dim DTS As DataSet = Dataset(Source)
            If DTS Is Nothing Then Return Nothing
            Dim SourceDataMember As String = Source.DataMember
            If DTS.Relations.Contains(SourceDataMember) Then
                Return DTS.Relations(SourceDataMember).ChildTable
            ElseIf DTS.Tables.Contains(SourceDataMember) Then
                Return DTS.Tables(SourceDataMember)
            Else
                Throw New Exception(String.Concat("Im Dataset anhand des BindingSource-Datamembers '",
                 SourceDataMember, "' keine Tabelle gefunden"))
            End If
        End Function

        <Extension()>
        Public Function All(Of T As DataRow)(subj As BindingSource, Optional filter As Func(Of T, Boolean) = Nothing) As IEnumerable(Of T)
            If filter.Null Then Return From drv In subj.Cast(Of DataRowView)() Select DirectCast(drv.Row, T)
            Return From drv In subj.Cast(Of DataRowView)() Where filter(DirectCast(drv.Row, T)) Select DirectCast(drv.Row, T)
        End Function

        ''' <summary> returnt die typisierte Datarow an aktueller Position - oder Nothing. </summary>
        <Extension()>
        Public Function At(Of T As DataRow)(subj As BindingSource) As T
            Return DirectCast(subj.At(subj.Position), T)
        End Function

        ''' <summary>
        ''' returnt die typisierte Datarow am index. Bei ungültigem index Nothing (keine OutOfRange-Exception!)
        ''' </summary>
        <Extension()>
        Public Function At(Of T As DataRow)(subj As BindingSource, index As Integer) As T
            Return DirectCast(subj.At(index), T)
        End Function

        ''' <summary> returnt die untypisierte Datarow an aktueller Position. </summary>
        <Extension()>
        Public Function At(subj As BindingSource) As DataRow
            Return subj.At(subj.Position)
        End Function

        ''' <summary>
        ''' returnt die Datarow am index. Bei ungültigem index Nothing (keine OutOfRange-Exception!)
        ''' </summary>
        <Extension()>
        Public Function At(subj As BindingSource, index As Integer) As DataRow
            If index < 0 OrElse index >= subj.Count Then Return Nothing
            Return DirectCast(subj(index), DataRowView).Row
        End Function

        <Extension()>
        Public Function FindX(Of T As DataRow)(subj As BindingSource, predicate As Func(Of T, Boolean)) As Integer
            For i As Integer = 0 To subj.Count - 1
                If predicate(DirectCast(DirectCast(subj(i), DataRowView).Row, T)) Then Return i
            Next
            Return -1
        End Function

        <Extension()>
        Public Function FindX(Of T)(subj As BindingSource, columnName As String, Key As T) As Integer
            'BindingSource.Find funzt nicht bei relateten BindingSourcses
            'FindX = subj.Find(PropertyName, Key)
            For i As Integer = 0 To subj.Count - 1
                If DirectCast(DirectCast(subj(i), DataRowView)(columnName), T).Equals(Key) Then Return i
            Next
            Return -1
        End Function

        Private Function KeysMatch(row As DataRow, cols As DataColumn(), keys As Object()) As Boolean
            For ii = 0 To cols.Length - 1
                If Not Object.Equals(keys(ii), row(cols(ii))) Then Return False
            Next
            Return True
        End Function

        ''' <summary> stellt den Datensatz mit den angegebenen Primärschlüssel-Werten ein </summary>
        <Extension()>
        Public Function MoveTo(subj As BindingSource, primKeys() As Object) As Boolean
            If subj.Count = 0 Then Return False
            Dim rw = DirectCast(subj(0), DataRowView).Row
            Dim cols = rw.Table.PrimaryKey
            Dim i = 0
            Try
                If KeysMatch(rw, cols, primKeys) Then Return True
                For i = 1 To subj.Count - 1
                    rw = DirectCast(subj(i), DataRowView).Row
                    If KeysMatch(rw, cols, primKeys) Then Return True
                Next
            Finally
                subj.Position = i
            End Try
            Return False
        End Function

        '''' <summary> stellt den Datensatz mit dem angegebenen Primärschlüssel-Wert ein </summary>
        <Extension()>
        Public Function MoveTo(Of T)(subj As BindingSource, primKey As T) As Boolean
            If subj.Count = 0 Then Return False
            Return subj.MoveTo(DirectCast(subj(0), DataRowView).Row.Table.PrimaryKey(0).ColumnName, primKey)
        End Function

        ''' <summary> 
        ''' stellt den ersten Datensatz ein, dessen Wert in der angegebenen spalte mit Key übereinstimmt
        ''' </summary>
        <Extension()>
        Public Function MoveTo(Of T)(subj As BindingSource, columnName As String, Key As T) As Boolean
            Dim i = subj.FindX(columnName, Key)
            MoveTo = i >= 0
            subj.Position = i
        End Function


        ''' <summary> stellt die erste matchende DataRow ein </summary>
        <Extension()>
        Public Function MoveToRow(Of T As DataRow)(subj As BindingSource, predicate As Func(Of T, Boolean)) As Boolean
            Dim dv = DirectCast(subj.List, DataView)
            For i As Integer = 0 To subj.Count - 1
                If predicate(DirectCast(dv(i).Row, T)) Then
                    subj.Position = i
                    Return True
                End If
            Next
            Return False
        End Function

        '''' <summary> stellt die angegebene DataRow ein </summary>
        <Extension()>
        Public Function MoveToRow(bs As BindingSource, row As DataRow) As Boolean
            If row Is bs.At Then Return True
            If bs.IsBindingSuspended Then Return False
            Dim dv = DirectCast(bs.List, DataView)
            For i As Integer = 0 To bs.Count - 1
                If dv(i).Row Is row Then
                    bs.CancelEdit() 'ansonsten setzter u.U. beim Moven die vorherige Row auf Rowstate.Changed
                    bs.Position = i
                    Return True
                End If
            Next
            Return False
        End Function

        ''' <summary> for debugging-purposes </summary>
        Public Function CheckChanges(dts As DataSet) As Boolean
            dts = dts.GetChanges
            If dts.Null Then MessageBox.Show("no Changes") : Return False
            Dim tbls = dts.Tables.Cast(Of DataTable)().ToArray
            Dim infos = From tbl In tbls From rw In tbl.Rows.Cast(Of DataRow)()
                        Select rw.GetType.Name & String.Concat(rw.ItemArray.Select(Function(itm) " " & itm.ToString))
            MessageBox.Show(String.Join(Lf, infos))
            Return True
        End Function

    End Module

End Namespace
